==> zino-0.9.8.2.tar.gz: zino-0.9.8.2/tpoller/tpoller.tcl <==
#
# $Id: tpoller.tcl,v 1.19 2009/04/02 08:45:06 he Exp $
#

#
# Traffic statistics poller
#

namespace import ::polljob::name
namespace import ::just::sparseTableWalk
namespace import ::just::asyncSparseTableWalk
namespace import ::config::conf*

#
# This is _really_ ugly.  Ensure that rfc2863.mib is loaded prior to
# rfc1213.mib, and likewise ensure that IANAifType.tc is loaded as early as
# possible.
#
set x [lsearch -exact $tnm(mibs:core) rfc1213.mib]
if { $x >= 0 } {
    set tnm(mibs:core) [linsert $tnm(mibs:core) $x rfc2863.mib]
}

set x [lsearch -exact $tnm(mibs:core) compat.tc]
if { $x >= 0 } {
    set tnm(mibs:core) [lreplace $tnm(mibs:core) $x $x compat.tc IANAifType.tc]
}

mib load rfc2863.mib
mib load IANAifType.tc
mib load cisco.mib


proc initDev { sh } {
}

proc destroyDev { sh } {

    forgetConfig $sh
}

proc pollDev { sh } {
    global errorInfo

    set dev [name $sh]
    # First check if we're to collect stats from this box
    if { [conf $dev "statistics"] != "yes" } {	return; }

#    puts [format "polling %s" [name $sh]]
    if [catch {
	# determine reachability and whether it
	# has rebooted/crashed since the last go-around
	uptimeCheck $sh
	# need to do this for each round
	# since interfaces may have been turned
	# on or cards can be hot-swapped
	getConfig $sh
	# do the walk...
	pollIntfs $sh
    } msg] {
	if [regexp "FAILED" $msg] { return; }
	puts stderr $errorInfo
    }
}

#
# Uptime checking
#

proc uptimeCheck { sh } {
    global cx

    set uptime [getUptime $sh]
    set name [name $sh]
    if { ! [expectedUptime $name $uptime] } {
	# record timestamp for use later
	set cx($name,reload) [expr [clock seconds] - $uptime / 100]
    }
    set cx($name,lastpoll) [clock seconds]
}

proc expectedUptime { name newup } {
    global cx uptimeSlop

    if { ! [info exists cx($name,uptime)] } { return 0; }
    set lastuptime $cx($name,uptime)
    set cx($name,uptime) $newup

    # We don't know what to expect in this case
    if { ! [info exists cx($name,lastpoll)] } { return 0; }

    # Estimate new expected uptime
    set now [clock seconds]
    set estup [expr $lastuptime + \
		   100 * ( $now - $cx($name,lastpoll) )]

    # 32-bit unsigned counter overflow (or too close), ignore;
    # uptime counter wraparound has happened or is imminent
    if { $estup > 0xffffffff - $uptimeSlop } {
	set msg [format \
"%s: uptime wraparound (or close), estup %s lastuptime %s lastpoll %s time %s" \
		     $name $estup $lastuptime \
		     $cx($name,lastpoll) $now]
	logerr $msg
	return 1
    }

    if { $cx($name,uptime) == 0 } { return 0; }

    if { $newup < $estup - $uptimeSlop } {
	set msg [format \
"%s: unexpected uptime, newup %s estup %s lastuptime %s lastpoll %s time %s" \
		     $name $newup $estup $uptimeSlop \
		     $lastuptime $cx($name,lastpoll) $now]
	logerr $msg
	return 0
    }

    # Everything as expected
    return 1
}

proc getUptime { sh } {

    if { [catch {set vbl [$sh get "sysUpTime.0"]} msg] } {
	set msg [format "Uptime poll FAILED for %s: %s" [name $sh] $msg]
	logerr $msg
	error $msg
    }
    set name [lindex [lindex $vbl 0] 0]
    set value [lindex [lindex $vbl 0] 2]
    return [mib scan $name $value]
}

#
#
#

# Need to check config on each round, since interface
# cards can nowadays be hot-swapped

proc getConfig { sh } {
    global cx

    set name [name $sh]
    set cx($name,sysid) [getSysObjID $sh]
    forgetConfig $sh
    if [catch {getAddrList $sh} msg] {
	set msg [format "Could not get AddrList from %s: %s" $name $msg]
	logerr $msg
	return
    }
    if [catch {getIfList $sh} msg] {
	set msg [format "Could not get IfList from %s: %s" $name $msg]
	logerr $msg
	return
    }
}

proc getSysObjID { sh } {
    if [catch {set vbl [$sh get "sysObjectID.0"]} msg] {
	set msg [format "sysObjectID poll FAILED for %s: %s" [name $sh] $msg]
	logerr $msg
	error $msg
    }
    set name  [lindex [lindex $vbl 0] 0]
    set value [lindex [lindex $vbl 0] 2]
    return [mib scan $name $value]
}

proc forgetConfig { sh } {
    global cx

    set name [name $sh]
    if [info exists cx($name,ifindexes)] {
	foreach ix $cx($name,ifindexes) {
	    catch {unset cx($name,address,$ix)}
	    catch {unset cx($name,type,$ix)}
	    catch {unset cx($name,speed,$ix)}
	    catch {unset cx($name,admin,$ix)}
	    catch {unset cx($name,descr,$ix)}
	}
	unset cx($name,ifindexes)
    }
}

proc getAddrList { sh } {

    $sh walk x "ipAdEntIfIndex ipAdEntAddr" \
	    { recordAddr $sh $x }
}

proc recordAddr { sh vbl } {
    global cx

    set ix [lindex [lindex $vbl 0] 2]
    set ad [lindex [lindex $vbl 1] 2]

    set name [name $sh]
    if { ! [info exists cx($name,address,$ix)] } {
	# only remember lowest address if interface
	# has more than one
	set cx($name,address,$ix) $ad

#	puts [format "New address: %s for %s,%s" $ad $name $ix]
    }
#	puts [format "Address already seen for %s,%s: %s %s (new)" \
#		$name $ix $cx($name,address,$ix) $ad]
#
}

proc getIfList { sh } {

    set name [name $sh]

    set vl "ifIndex ifType ifSpeed ifAdminStatus ifDescr"
    if { [::config::conf $name "hcounters"] != "no" } {
	lappend vl "ifHighSpeed"
    }
    if [catch {
	# This table may be sparsely populated
	sparseTableWalk $sh x $vl { recordIntf $sh $x }
    } msg] {
	set msg [format "getIfList FAILED for %s: %s" [name $sh] $msg]
	logerr $msg
	error $msg
    }
}

proc recordIntf { sh vbl } {
    global cx

    set ix	[lindex [lindex $vbl 0] 2]
    set type	[lindex [lindex $vbl 1] 2]
    set speed	[lindex [lindex $vbl 2] 2]
    set admin	[lindex [lindex $vbl 3] 2]
    set descr	[lindex [lindex $vbl 4] 2]
    set type [mib scan [lindex [lindex $vbl 1] 0] $type]; # undo TC coding
    set hs	[lindex [lindex $vbl 5] 2]

    if { $hs != "" && $hs != 0 } {
	set speed [expr $hs * 1000000.0]
    }

    set name [name $sh]

    set cx($name,type,$ix) $type
    set cx($name,speed,$ix) $speed
    set cx($name,admin,$ix) $admin
    set cx($name,descr,$ix) $descr
    lappend cx($name,ifindexes) $ix

#    puts [format "Intf %s,%s type %s speed %s admin %s descr %s" \
#	    $name $ix $type $speed $admin $descr]

}

#
#
#

proc pollIntfs { sh } {

    set vl [getVarList $sh]
    asyncSparseTableWalk $sh x $vl {
	logVars %S $x
    } {
	# no endscript
#	puts [format "finished polling %s" [name %S]]
    } {
	set msg [format "%s poll FAILED: %s" [name %S] %E]
	logerr $msg
#	error $msg
    }
}

proc getVarList { sh } {
    global cx

    set name [::polljob::name $sh]

    set vl [list "ifInUcastPkts" "ifInNUcastPkts" "ifInDiscards" \
	    "ifInErrors" "ifOutUcastPkts" "ifOutNUcastPkts" \
	    "ifOutDiscards" "ifOutErrors"]

    if { [::config::conf $name "hcounters"] != "no" } {
	lappend vl "ifHCInOctets" "ifHCOutOctets" "ifInOctets" "ifOutOctets"
    } else {
	lappend vl "ifInOctets" "ifOutOctets"
    }

    if { [oidPfxMatch [mib oid cisco] $cx([name $sh],sysid)] } {
	lappend vl "locIfInRunts" "locIfInGiants" \
		"locIfInCRC" "locIfInFrame" \
		"locIfInOverrun" "locIfInIgnored" \
		"locIfInAbort" "locIfResets" \
		"locIfRestarts" "locIfCollisions"
    }
    return $vl
}

proc oidPfxMatch { a b } {
    append a "."
    set ix [string first $a $b]
    return [expr $ix == 0]
}

proc logVars { sh vbl } {
    global cx

    set stamp [getNowStamp]
    set name [name $sh]

    foreach vb $vbl {
	set oid [lindex $vb 0]
	if { $oid != {} } {
	    set ix [getInstance $oid]
	}
    }
    if { ! [info exists ix] } { return; }

    if { ! [info exists cx($name,descr,$ix)] &&
	 ! [info exists cx($name,admin,$ix)] } {
	return;				# mostly-non-existent ifIndex
	# old Cisco MIB might return vars for locIf*
    }

    if [confExists $name "watchpat"] {
	set wp [conf $name "watchpat"]
	# Only poll ports matching pattern
	if { ! [regexp $wp $cx($name,descr,$ix)] } { return; }
    }
    if [confExists $name "ignorepat"] {
	set ign [conf $name "ignorepat"]
	# Ignore ports we've been instructed to ignore
	if [regexp $ign $cx($name,descr,$ix)] { return; }
    }
    # Ignore interfaces not turned on
    if { $cx($name,admin,$ix) != "up" } { return; }

    set fh [openFile $name $ix newfile]
    # Log files rotated daily, so log interface
    # information first and close to the end of the day
    if { $newfile || [almostMidnight]} {
	logInfo $fh $name $ix
    }

    # if the device was reloaded
    # and we've not yet recorded it, do so now
    maybeLogReload $fh $name $ix

    foreach vb $vbl {
	set oid [lindex $vb 0]
	# skip nonexistent table entries
	if { $oid == {} } { continue; }

	set vn [mib name $oid]
	puts $fh [format "%s %s %s" $stamp $vn [lindex $vb 2]]
    }

    close $fh
}

proc getInstance { oid } {
    return [join [lrange [split [mib name $oid] "."] 1 end] "."]
}

proc openFile { dev ix new } {
    upvar $new New

    set fn [getFileName $dev $ix]
    set New [expr ! [file exists $fn]]
    if [catch {set fh [open $fn "a"]} msg] {
	logerr [format "%s open error: %s" $fn $msg]
	error $msg
    }
    return $fh
}

proc getFileName { dev ix } {
    global cx LogDir

    if { [info exists cx($dev,address,$ix)] } {
	set aPart $cx($dev,address,$ix)
    } else {
	set aPart [format "#%d" $ix]
    }
    return [format "%s/%s.%s" $LogDir $dev $aPart]
}

proc almostMidnight {} {

    set s [clock seconds]
    set h [clock format $s -format "%H"]
    set m [clock format $s -format "%M"]
    if { $h == 23 && $m >= 35 } {
	return 1
    }
    return 0
}

proc maybeLogReload { fh name ix } {
    global cx

    if [info exists cx($name,reload)] {
	if [info exists cx($name,reload,$ix)] {
	    if { $cx($name,reload) != $cx($name,reload,$ix) } {
		logReload $fh $name
		set cx($name,reload,$ix) $cx($name,reload); # mark as logged
	    }
	} else {
	    logReload $fh $name
	    set cx($name,reload,$ix) $cx($name,reload)
	}
    }
}

proc logReload { fh name } {
    global cx

    puts $fh [format "%s Reboot at %s" \
	    [getNowStamp] [getStamp $cx($name,reload)]]
}

proc logInfo { fh name ix } {
    global cx

    puts $fh [format "ifSpeed %s" $cx($name,speed,$ix)]
    puts $fh [format "ifType %d"  $cx($name,type,$ix)]
    puts $fh [format "ifDescr \"%s\"" $cx($name,descr,$ix)]
}

#
#
#

proc getNowStamp {} {
    return [getStamp [clock seconds]]
}

proc getStamp { secs } {
    return [clock format $secs \
	    -format "%a %b %d %H:%M:%S %Y"]
}

proc logerr { msg } {

    # Could / should be improved to do syslog (?)
    puts stderr [format "%s: %s" [getNowStamp] $msg]
}

#
#
#

==> zino-0.9.9.4.tar.gz: zino-0.9.9.4/tpoller/tpoller.tcl <==
#
# $Id: tpoller.tcl,v 1.22 2010-09-24 12:29:17 he Exp $
#

#
# Traffic statistics poller
#

namespace import ::polljob::name
namespace import ::just::sparseTableWalk
namespace import ::just::asyncSparseTableWalk
namespace import ::config::conf*

#
# This is _really_ ugly.  Ensure that rfc2863.mib is loaded prior to
# rfc1213.mib, and likewise ensure that IANAifType.tc is loaded as early as
# possible.
#
set x [lsearch -exact $tnm(mibs:core) rfc1213.mib]
if { $x >= 0 } {
    set tnm(mibs:core) [linsert $tnm(mibs:core) $x rfc2863.mib]
}

set x [lsearch -exact $tnm(mibs:core) compat.tc]
if { $x >= 0 } {
    set tnm(mibs:core) [lreplace $tnm(mibs:core) $x $x compat.tc IANAifType.tc]
}

mib load rfc2863.mib
mib load IANAifType.tc
mib load cisco.mib


proc initDev { sh } {
}

proc destroyDev { sh } {

    forgetConfig $sh
}

proc pollDev { sh } {
    global errorInfo

    set dev [name $sh]
    # First check if we're to collect stats from this box
    if { [conf $dev "statistics"] != "yes" } {	return; }

#    puts [format "polling %s" [name $sh]]

    $sh get sysUpTime.0 {
	uptimeProcess "%S" "%E" "%V" pollDev_1
    }
}

proc pollDev_1 { sh } {
    
    $sh get sysObjectID.0 {
	sysobjProcess "%S" "%E" "%V" pollDev_2
    }
}

proc pollDev_2 { sh } {

    forgetConfig $sh
    bulkWalk $sh "ipAdEntIfIndex ipAdEntAddr" 10 \
	recordAddr pollDev_3 errAddr
}

proc pollDev_3 { sh } {

    set name [name $sh]

    set vl "ifIndex ifType ifSpeed ifAdminStatus ifDescr"
    if { [::config::conf $name "hcounters"] != "no" } {
	lappend vl "ifHighSpeed"
    }
    asyncSparseTableWalk $sh x $vl { recordIntf "%S" $x } \
	{ pollDev_4 "%S" } { errIntf "%S" "%E" }
}

proc pollDev_4 { sh } {

    pollIntfs $sh
}

proc errAddr { sh err } {

    set msg [format "Get address list FAILED for %s: %s" [name $sh] $err]
    logerr $msg
}

proc errIntf { sh err } {

    set msg [format "Get interface list FAILED for %s: %s" [name $sh] $err]
    logerr $msg
}

#
# Uptime checking
#

proc uptimeProcess { sh err vbl call_continue } {

    if { $err != "noError" } {
	set msg [format "Uptime poll FAILED for %s: %s" [name $sh] $err]
	logerr $msg
	return
    }
    set name [lindex [lindex $vbl 0] 0]
    set value [lindex [lindex $vbl 0] 2]
    set uptime [mib scan $name $value]
    if { ! [expectedUptime $name $uptime] } {
	# record timestamp for use later
	set cx($name,reload) [expr [clock seconds] - $uptime / 100]
    }
    set cx($name,lastpoll) [clock seconds]

    eval $call_continue $sh
}

proc expectedUptime { name newup } {
    global cx uptimeSlop

    if { ! [info exists cx($name,uptime)] } { return 0; }
    set lastuptime $cx($name,uptime)
    set cx($name,uptime) $newup

    # We don't know what to expect in this case
    if { ! [info exists cx($name,lastpoll)] } { return 0; }

    # Estimate new expected uptime
    set now [clock seconds]
    set estup [expr $lastuptime + \
		   100 * ( $now - $cx($name,lastpoll) )]

    # 32-bit unsigned counter overflow (or too close), ignore;
    # uptime counter wraparound has happened or is imminent
    if { $estup > 0xffffffff - $uptimeSlop } {
	set msg [format \
"%s: uptime wraparound (or close), estup %s lastuptime %s lastpoll %s time %s" \
		     $name $estup $lastuptime \
		     $cx($name,lastpoll) $now]
	logerr $msg
	return 1
    }

    if { $cx($name,uptime) == 0 } { return 0; }

    if { $newup < $estup - $uptimeSlop } {
	set msg [format \
"%s: unexpected uptime, newup %s estup %s lastuptime %s lastpoll %s time %s" \
		     $name $newup $estup $uptimeSlop \
		     $lastuptime $cx($name,lastpoll) $now]
	logerr $msg
	return 0
    }

    # Everything as expected
    return 1
}

#
#
#

# Need to check config on each round, since interface
# cards can nowadays be hot-swapped

proc sysobjProcess { sh err vbl call_continue } {
    global cx

    if { $err != "noError" } {
	set msg [format "sysObjectID poll FAILED for %s: %s" [name $sh] $err]
	logerr $msg
	return
    }
    set name  [lindex [lindex $vbl 0] 0]
    set value [lindex [lindex $vbl 0] 2]
    set sysobj [mib scan $name $value]

    set cx([name $sh],sysid) $sysobj

    eval $call_continue $sh
}

proc forgetConfig { sh } {
    global cx

    set name [name $sh]
    if [info exists cx($name,ifindexes)] {
	foreach ix $cx($name,ifindexes) {
	    catch {unset cx($name,address,$ix)}
	    catch {unset cx($name,type,$ix)}
	    catch {unset cx($name,speed,$ix)}
	    catch {unset cx($name,admin,$ix)}
	    catch {unset cx($name,descr,$ix)}
	}
	unset cx($name,ifindexes)
    }
}

proc recordAddr { sh vbl } {
    global cx

    set ix [lindex [lindex $vbl 0] 2]
    set ad [lindex [lindex $vbl 1] 2]

    set name [name $sh]
    if { ! [info exists cx($name,address,$ix)] } {
	# only remember lowest address if interface
	# has more than one
	set cx($name,address,$ix) $ad

#	puts [format "New address: %s for %s,%s" $ad $name $ix]
    }
#	puts [format "Address already seen for %s,%s: %s %s (new)" \
#		$name $ix $cx($name,address,$ix) $ad]
#
}

proc recordIntf { sh vbl } {
    global cx

    set ix	[lindex [lindex $vbl 0] 2]
    set type	[lindex [lindex $vbl 1] 2]
    set speed	[lindex [lindex $vbl 2] 2]
    set admin	[lindex [lindex $vbl 3] 2]
    set descr	[lindex [lindex $vbl 4] 2]
    set type [mib scan [lindex [lindex $vbl 1] 0] $type]; # undo TC coding
    set hs	[lindex [lindex $vbl 5] 2]

    if { $hs != "" && $hs != 0 } {
	set speed [expr $hs * 1000000.0]
    }

    set name [name $sh]

    set cx($name,type,$ix) $type
    set cx($name,speed,$ix) $speed
    set cx($name,admin,$ix) $admin
    set cx($name,descr,$ix) $descr
    lappend cx($name,ifindexes) $ix

#    puts [format "Intf %s,%s type %s speed %s admin %s descr %s" \
#	    $name $ix $type $speed $admin $descr]
}

#
#
#

proc pollIntfs { sh } {

    set vl [getVarList $sh]
    asyncSparseTableWalk $sh x $vl {
	logVars %S $x
    } {
	# no endscript
#	puts [format "finished polling %s" [name %S]]
    } {
	logerr [format "%s poll FAILED: %s" [name %S] %E]
    }
}

proc getVarList { sh } {
    global cx

    set name [::polljob::name $sh]

    set vl [list "ifInUcastPkts" "ifInNUcastPkts" "ifInDiscards" \
	    "ifInErrors" "ifOutUcastPkts" "ifOutNUcastPkts" \
	    "ifOutDiscards" "ifOutErrors"]

    if { [::config::conf $name "hcounters"] != "no" } {
	lappend vl "ifHCInOctets" "ifHCOutOctets" "ifInOctets" "ifOutOctets"
    } else {
	lappend vl "ifInOctets" "ifOutOctets"
    }

    if { [oidPfxMatch [mib oid cisco] $cx([name $sh],sysid)] } {
	lappend vl "locIfInRunts" "locIfInGiants" \
		"locIfInCRC" "locIfInFrame" \
		"locIfInOverrun" "locIfInIgnored" \
		"locIfInAbort" "locIfResets" \
		"locIfRestarts" "locIfCollisions"
    }
    return $vl
}

proc oidPfxMatch { a b } {
    append a "."
    set ix [string first $a $b]
    return [expr $ix == 0]
}

proc logVars { sh vbl } {
    global cx

    set stamp [getNowStamp]
    set name [name $sh]

    foreach vb $vbl {
	set oid [lindex $vb 0]
	if { $oid != {} } {
	    set ix [getInstance $oid]
	}
    }
    if { ! [info exists ix] } { return; }

    if { ! [info exists cx($name,descr,$ix)] &&
	 ! [info exists cx($name,admin,$ix)] } {
	return;				# mostly-non-existent ifIndex
	# old Cisco MIB might return vars for locIf*
    }

    if [confExists $name "watchpat"] {
	set wp [conf $name "watchpat"]
	# Only poll ports matching pattern
	if { ! [regexp $wp $cx($name,descr,$ix)] } { return; }
    }
    if [confExists $name "ignorepat"] {
	set ign [conf $name "ignorepat"]
	# Ignore ports we've been instructed to ignore
	if [regexp $ign $cx($name,descr,$ix)] { return; }
    }
    # Ignore interfaces not turned on
    if { $cx($name,admin,$ix) != "up" } { return; }

    set fh [openFile $name $ix newfile]
    # Log files rotated daily, so log interface
    # information first and close to the end of the day
    if { $newfile || [almostMidnight]} {
	logInfo $fh $name $ix
    }

    # if the device was reloaded
    # and we've not yet recorded it, do so now
    maybeLogReload $fh $name $ix

    foreach vb $vbl {
	set oid [lindex $vb 0]
	# skip nonexistent table entries
	if { $oid == {} } { continue; }

	set vn [mib name $oid]
	puts $fh [format "%s %s %s" $stamp $vn [lindex $vb 2]]
    }

    close $fh
}

proc getInstance { oid } {
    return [join [lrange [split [mib name $oid] "."] 1 end] "."]
}

proc openFile { dev ix new } {
    upvar $new New

    set fn [getFileName $dev $ix]
    set New [expr ! [file exists $fn]]
    if [catch {set fh [open $fn "a"]} msg] {
	logerr [format "%s open error: %s" $fn $msg]
	error $msg
    }
    return $fh
}

proc getFileName { dev ix } {
    global cx LogDir

    if { [info exists cx($dev,address,$ix)] } {
	set aPart $cx($dev,address,$ix)
    } else {
	set aPart [format "#%d" $ix]
    }
    return [format "%s/%s.%s" $LogDir $dev $aPart]
}

proc almostMidnight {} {

    set s [clock seconds]
    set h [clock format $s -format "%H"]
    set m [clock format $s -format "%M"]
    if { $h == 23 && $m >= 35 } {
	return 1
    }
    return 0
}

proc maybeLogReload { fh name ix } {
    global cx

    if [info exists cx($name,reload)] {
	if [info exists cx($name,reload,$ix)] {
	    if { $cx($name,reload) != $cx($name,reload,$ix) } {
		logReload $fh $name
		set cx($name,reload,$ix) $cx($name,reload); # mark as logged
	    }
	} else {
	    logReload $fh $name
	    set cx($name,reload,$ix) $cx($name,reload)
	}
    }
}

proc logReload { fh name } {
    global cx

    puts $fh [format "%s Reboot at %s" \
	    [getNowStamp] $cx($name,reload)]
}

proc logInfo { fh name ix } {
    global cx

    puts $fh [format "Version 1"]
    puts $fh [format "ifSpeed %s" $cx($name,speed,$ix)]
    puts $fh [format "ifType %d"  $cx($name,type,$ix)]
    puts $fh [format "ifDescr \"%s\"" $cx($name,descr,$ix)]
}

#
#
#

proc getNowStamp {} {
    return [clock seconds]
}

proc dateNowStamp { } {
    return [clock format [clock seconds] \
	    -format "%a %b %d %H:%M:%S %Y"]
    
}

proc logerr { msg } {

    # Could / should be improved to do syslog (?)
    puts stderr [format "%s: %s" [dateNowStamp] $msg]
}

#
#
#
